"""Public init."""

from .queue import Queue

__all__ = ["Queue"]

# NOTE: `__version__` is not defined because this package is built using 'setuptools-scm' --
#   use `importlib.metadata.version(...)` if you need to access version info at runtime.
